/*****************************************************************************************************
    Ady Bloom Revision 5 rewritten for ReShade

    Copyright 2019 Adyss

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*****************************************************************************************************/

#include "ReShade.fxh"
#include "ReShadeUI.fxh"

////////////////////////////
// UI /////////////////////

uniform bool TonemapToggle <
	ui_label = "Enable Tonemap";
	ui_tooltip = "Tonemaps the bloom ";
    ui_category = "Debug";
> = false;

uniform bool LensDirtToggle <
	ui_label = "Enable Lensdirt";
	ui_tooltip = "Adds Dirt to the Lens ";
    ui_category = "Lensdirt";
> = false;

uniform float3 DirtTint <
    ui_label = "Lensdirt Tint";
    ui_tooltip = "Tints the Lensdirt";
    ui_category = "Lensdirt";
    ui_type = "color";
    ui_min = 0.0;
    ui_max = 1.0;
> = float3(1.0, 1.0, 1.0);

uniform float DirtPower <
    ui_label    = "Lensdirt Power";
    ui_tooltip  = "Defines how visible the Lensdirt is";
    ui_category = "Lensdirt";
    ui_type     = "slider";
    ui_min      = 0.0;
    ui_max      = 3.0;
> = 0.5;

uniform float Threshold <
    ui_label    = "Bloom Threshold";
    ui_tooltip  = "Defines what triggers bloom by filtering darker objects";
    ui_category = "Prepass";
    ui_type     = "slider";
    ui_min      = 0.0;
    ui_max      = 1.0;
> = 0.1;

uniform float BloomSensitivity <
    ui_label    = "Bloom Sensitivity";
    ui_tooltip  = "A Curve that is applied to the bloom input";
    ui_category = "Prepass";
    ui_type     = "slider";
    ui_min      = 0.1;
    ui_max      = 5.0;
> = 1.0;

uniform float BloomSaturation <
    ui_label    = "Bloom Saturation";
    ui_tooltip  = "Makes the bloom appear more satuated";
    ui_category = "Prepass";
    ui_type     = "slider";
    ui_min      = 0.0;
    ui_max      = 3.0;
> = 1.0;

uniform float BloonPower <
    ui_label    = "Bloom Power";
    ui_tooltip  = "Controls how strong the bloom is";
    ui_category = "Prepass";
    ui_type     = "slider";
    ui_min      = 0.0;
    ui_max      = 10.0;
> = 1.0;

uniform float LumaPreservingBlend <
    ui_label    = "Luma Preservation";
    ui_tooltip  = "How much bloom will light up shadows";
    ui_category = "Prepass";
    ui_type     = "slider";
    ui_min      = 0.0;
    ui_max      = 1.0;
> = 0.8;

uniform int BloomSamples <
    ui_label    = "Bloom Samples";
    ui_tooltip  = "Defines how many blur taps are used. This also means the radius gets bigger";
    ui_category = "Bloom Kernel";
    ui_type     = "slider";
    ui_min      = 4.0;
    ui_max      = 32.0;
> = 8.0;

uniform float BloomCurve <
    ui_label    = "Bloom Curve";
    ui_tooltip  = "Defines the way the bloom spreads";
    ui_category = "Bloom Kernel";
    ui_type     = "slider";
    ui_min      = 0.1;
    ui_max      = 5.0;
> = 1.0;

uniform float SigmaH <
    ui_label    = "Horizontal Sigma";
    ui_tooltip  = "How big the bloom is Horizontally";
    ui_category = "Bloom Kernel";
    ui_type     = "slider";
    ui_min      = 0.2;
    ui_max      = 5.0;
> = 1.0;

uniform float SigmaV <
    ui_label    = "Vertical Sigma";
    ui_tooltip  = "How big the bloom is Vertically";
    ui_category = "Bloom Kernel";
    ui_type     = "slider";
    ui_min      = 0.2;
    ui_max      = 5.0;
> = 1.0;

////////////////////////////
// Textures ///////////////

texture2D BloomTex1 	        	{ Width = BUFFER_WIDTH / 2; 	Height = BUFFER_HEIGHT / 2;    Format = RGBA16F;};
texture2D BloomTex2		         	{ Width = BUFFER_WIDTH / 4;  	Height = BUFFER_HEIGHT / 4;    Format = RGBA16F;};
texture2D BloomTex3 	        	{ Width = BUFFER_WIDTH / 8;  	Height = BUFFER_HEIGHT / 8;    Format = RGBA16F;};
texture2D BloomTex4         		{ Width = BUFFER_WIDTH / 16;    Height = BUFFER_HEIGHT / 16;   Format = RGBA16F;};
texture2D BloomTex5 	        	{ Width = BUFFER_WIDTH / 32;    Height = BUFFER_HEIGHT / 32;   Format = RGBA16F;};
texture2D BloomTex6         		{ Width = BUFFER_WIDTH / 64;    Height = BUFFER_HEIGHT / 64;   Format = RGBA16F;};
texture2D BloomTex7         		{ Width = BUFFER_WIDTH / 128;   Height = BUFFER_HEIGHT / 128;  Format = RGBA16F;};
texture2D OriginalTexture           { Width = BUFFER_WIDTH;         Height = BUFFER_HEIGHT;        Format = RGBA8;};
texture2D DirtTex  <source="MagicBloom_Dirt.png";> { Width = BUFFER_WIDTH; Height = BUFFER_HEIGHT; };

sampler2D TextureBloom1     		{ Texture = BloomTex1;};
sampler2D TextureBloom2	         	{ Texture = BloomTex2;};
sampler2D TextureBloom3	        	{ Texture = BloomTex3;};
sampler2D TextureBloom4     		{ Texture = BloomTex4;};
sampler2D TextureBloom5	        	{ Texture = BloomTex5;};
sampler2D TextureBloom6     		{ Texture = BloomTex6;};
sampler2D TextureBloom7	        	{ Texture = BloomTex7;};
sampler2D OriginalColor	        	{ Texture = OriginalTexture;};
sampler2D DirtSampler	        	{ Texture = DirtTex;};

////////////////////////////
// Functions //////////////

float4 SimpleBlur(sampler2D InputTex, float2 Coord, float2 pixelsize)
{
    float4 Blur = 0.0;

    static const float2 Offsets[4]=
    {
        float2(0.5, 0.5),
        float2(0.5, -0.5),
        float2(-0.5, 0.5),
        float2(-0.5, -0.5)
    };

    for (int i = 0; i < 4; i++)
    {
        Blur += tex2D(InputTex, Coord + Offsets[i] * pixelsize);
    }

    return Blur * 0.25;
}

struct BloomData
{
    float4 Blur;
    float2 Pixelsize;
    float2 Coord;
    float  Offset;
    float  Weight;
    float  WeightSum;
};

float2 GetPixelSize(float texsize)
{
    return (1 / texsize) * float2(1, BUFFER_WIDTH / BUFFER_HEIGHT);
}

// Get Luma function by TreyM
// Luma Coefficients
#define Rec709      0
#define Rec709_5    1
#define Rec601      2
#define Rec2020     3
#define Lum333      4

// Calculate perceived luminance color by using the ITU-R BT standards
float GetLuma(in float3 color, int btspec)
{
    static const float3 LumaCoeff[5] =
    {
        // 0: HD TV - Rec.709
        float3(0.2126, 0.7152f, 0.0722),
        // 1: HD TV - Rec.709-5
        float3(0.212395, 0.701049, 0.086556),
        // 2: CRT TV - Rec.601
        float3(0.299, 0.587, 0.114),
        // 3: HDR Spec - Rec.2020
        float3(0.2627, 0.6780, 0.0593),
        // 4: Incorrect Equal Weighting
        float3(0.3333, 0.3333, 0.3333)
    };

    //return sqrt(dot(color.rgb * color.rgb, LumaCoeff[btspec])); // Requires sRGB input
    return dot(color.rgb, LumaCoeff[btspec]);
}

////////////////////////////
// PixelShaders ///////////

float4 PS_Prepass(float4 pos : SV_Position, float2 coord : TEXCOORD) : SV_Target
{
    float4 Color    = tex2D(ReShade::BackBuffer, coord);

           // Gamma Curve
           Color    = pow(Color, BloomSensitivity);

           // Luma Threshold
    float  Luma     = GetLuma(Color, Rec709);
           Color   /= max(Luma, 0.001);
           Luma     = max(0.0, Luma - Threshold);
           Color   *= Luma;

           Color    = lerp(Luma, Color, BloomSaturation);

    return saturate(Color);
}

float4  BloomH(float2 Coord, float texsize, sampler2D InputTex)
{
    BloomData Bloom;
    Bloom.Pixelsize      = GetPixelSize(texsize);
    Bloom.Offset         = SigmaH * 0.5;

    for (int i = 1; i < BloomSamples; i++)
    {
        Bloom.Weight     = pow(BloomSamples - i, BloomCurve);
        Bloom.Blur      += tex2D(InputTex, Coord + (float2(Bloom.Offset, 0) * Bloom.Pixelsize)) * Bloom.Weight;
        Bloom.Blur      += tex2D(InputTex, Coord - (float2(Bloom.Offset, 0) * Bloom.Pixelsize)) * Bloom.Weight;
        Bloom.Offset    += SigmaH;
        Bloom.WeightSum += Bloom.Weight;
    }

    return Bloom.Blur /= Bloom.WeightSum * 2;
}

float4  BloomV(float2 Coord, float texsize, sampler2D InputTex)
{
    BloomData Bloom;
    Bloom.Pixelsize      = GetPixelSize(texsize);
    Bloom.Offset         = SigmaV * 0.5;

    for (int i = 1; i < BloomSamples; i++)
    {
        Bloom.Weight     = pow(BloomSamples - i, BloomCurve);
        Bloom.Blur      += tex2D(InputTex, Coord + (float2(0, Bloom.Offset) * Bloom.Pixelsize)) * Bloom.Weight;
        Bloom.Blur      += tex2D(InputTex, Coord - (float2(0, Bloom.Offset) * Bloom.Pixelsize)) * Bloom.Weight;
        Bloom.Offset    += SigmaV;
        Bloom.WeightSum += Bloom.Weight;
    }

    return Bloom.Blur /= Bloom.WeightSum * 2;
}


// Cant make this more compact in reshade... kinda sucks
float3 PS_StoreOriginal(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return tex2D(ReShade::BackBuffer, Coord);
}

float4 PS_BloomH1(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 2, TextureBloom1);
}

float4 PS_BloomV1(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 2, ReShade::BackBuffer);
}

float4 PS_BloomH2(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 4, TextureBloom1);
}

float4 PS_BloomV2(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 4, ReShade::BackBuffer);
}

float4 PS_BloomH3(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 8, TextureBloom2);
}

float4 PS_BloomV3(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 8, ReShade::BackBuffer);
}

float4 PS_BloomH4(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 16, TextureBloom3);
}

float4 PS_BloomV4(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 16, ReShade::BackBuffer);
}

float4 PS_BloomH5(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 32, TextureBloom4);
}

float4 PS_BloomV5(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 32, ReShade::BackBuffer);
}

float4 PS_BloomH6(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 64, TextureBloom5);
}

float4 PS_BloomV6(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 64, ReShade::BackBuffer);
}

float4 PS_BloomH7(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomH(Coord, BUFFER_WIDTH / 128, TextureBloom6);
}

float4 PS_BloomV7(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    return BloomV(Coord, BUFFER_WIDTH / 128, ReShade::BackBuffer);
}

float3 blend_screen(float3 a, float3 b) {
    return 1.0 - (1.0 - a) * (1.0 - b);
}


float3 PS_BloomMix(float4 pos : SV_Position, float2 Coord : TEXCOORD) : SV_Target
{
    float3 Color  = tex2D(OriginalColor, Coord).rgb;
    float3 Bloom  = SimpleBlur(TextureBloom1, Coord, GetPixelSize(BUFFER_WIDTH / 2)).rgb;
           Bloom += SimpleBlur(TextureBloom2, Coord, GetPixelSize(BUFFER_WIDTH / 4)).rgb;
           Bloom += SimpleBlur(TextureBloom3, Coord, GetPixelSize(BUFFER_WIDTH / 8)).rgb;
           Bloom += SimpleBlur(TextureBloom4, Coord, GetPixelSize(BUFFER_WIDTH / 16)).rgb;
           Bloom += SimpleBlur(TextureBloom5, Coord, GetPixelSize(BUFFER_WIDTH / 32)).rgb;
           Bloom += SimpleBlur(TextureBloom6, Coord, GetPixelSize(BUFFER_WIDTH / 64)).rgb;
           Bloom += SimpleBlur(TextureBloom7, Coord, GetPixelSize(BUFFER_WIDTH / 128)).rgb;
           Bloom *= BloonPower / 7;  // 1/7 = 0.1428571428571429

           if(TonemapToggle)
           Bloom  = Bloom / (Bloom + 1);

           if(LensDirtToggle)
           Bloom += (tex2D(DirtSampler, Coord) * Bloom) * DirtTint * DirtPower;

           Color += lerp(Bloom, max(Bloom - Color, 0.0), LumaPreservingBlend);

    return saturate(Color);
}

technique AdyBloom
{
    pass Bloom1
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_StoreOriginal;
        RenderTarget0 = OriginalTexture;
	}
	pass Bloom2
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_Prepass;
        RenderTarget0 = BloomTex1;
	}
	pass Bloom3
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH1;
	}
    pass Bloom4
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV1;
        RenderTarget0 = BloomTex1;
	}
    pass Bloom5
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH2;
	}
    pass Bloom6
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV2;
        RenderTarget0 = BloomTex2;
	}
    pass Bloom7
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH3;
	}
    pass Bloom8
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV3;
        RenderTarget0 = BloomTex3;
	}
    pass Bloom9
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH4;
	}
    pass Bloom10
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV4;
        RenderTarget0 = BloomTex4;
	}
    pass Bloom11
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH5;
	}
    pass Bloom12
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV5;
        RenderTarget0 = BloomTex5;
	}
    pass Bloom13
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH6;
	}
    pass Bloom14
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV6;
        RenderTarget0 = BloomTex6;
	}
    pass Bloom15
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomH7;
	}
    pass Bloom16
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomV7;
        RenderTarget0 = BloomTex7;
	}
    pass Bloom17
	{
		VertexShader  = PostProcessVS;
		PixelShader   = PS_BloomMix;
	}
}
