/*******************************************************************************
    Author: Vortigern

    License: MIT, Copyright (c) 2023 Vortigern

    MIT License

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#pragma once

// REST addov v1.1.0 has issues with hardware linearization
#ifndef USE_HW_LIN
    #define USE_HW_LIN 0
#endif

#ifndef V_ENABLE_BLOOM
    #define V_ENABLE_BLOOM 1
#endif

#ifndef V_ENABLE_SHARPEN
    #define V_ENABLE_SHARPEN 1
#endif

#ifndef V_ENABLE_COLOR_STUDIO
    #define V_ENABLE_COLOR_STUDIO 1
#endif

#ifndef V_BLOOM_MANUAL_PASSES
    #define V_BLOOM_MANUAL_PASSES 0 // if 0 -> auto select depending on the resolution, else -> 2 <= X <= 9
#endif

#ifndef V_BLOOM_DEBUG
    #define V_BLOOM_DEBUG 0
#endif

#if IS_SRGB
    #define CAT_TONEMAP "Tonemap"

    UI_BOOL(CAT_TONEMAP, UI_CC_ShowHDR, "Show HDR Colors", "If on, shows colors which are in HDR before tonemapping is applied.", false)
    UI_FLOAT(CAT_TONEMAP, UI_CC_TonemapMod, "Tonemap Modifier", "Lower values increase the color range, but can introduce clipping.", 1.001, 1.5, 1.025)
    UI_FLOAT(CAT_TONEMAP, UI_CC_TonemapExp, "Tonemap Exponent", "Increase to modify the tonemapping curve to a more filmic one.", 1.0, 3.0, 1.0)
#else
    #define CAT_HDR "HDR"

    UI_INT(CAT_HDR, UI_CC_HDRWhiteLvl, "HDR White Level", "Adjust if you have changed the game's HDR white point", 1, 1e4, 203)
#endif

#if V_ENABLE_BLOOM
    #define CAT_BLOOM "Bloom"

    UI_FLOAT(CAT_BLOOM, UI_Bloom_Intensity, "Bloom Intensity", "Controls the amount of bloom", 0.0, 1.0, 0.05)
    UI_FLOAT(CAT_BLOOM, UI_Bloom_Radius, "Bloom Radius", "Affects the size/scale of the bloom", 0.0, 1.0, 0.85)
    UI_FLOAT(CAT_BLOOM, UI_Bloom_NoiseStrength, "Noise Strength", "How much noise to add.", 0.0, 1.0, 0.25)
#endif

#if V_ENABLE_SHARPEN
    #define CAT_SHARP "Sharpen"

    UI_FLOAT(CAT_SHARP, UI_CC_SharpenLimit, "Sharpening Limit", "Limits the sharpening to prevent halos around objects", 0.0, 0.1, 0.01)
    UI_FLOAT(CAT_SHARP, UI_CC_SharpenStrength, "Sharpening Strength", "Controls the shaprening strength.", 0.0, 2.0, 1.0)
    UI_FLOAT(CAT_SHARP, UI_CC_UnsharpenStrength, "Unsharpening Strength", "Controls the unsharpness strength.", 0.0, 2.0, 1.0)
    UI_FLOAT(CAT_SHARP, UI_CC_SharpenDepthLimit, "Depth Limit", "Depth limit for the sharpening and unsharpening", 0.0, 1.0, 0.125)
#endif

#if V_ENABLE_COLOR_STUDIO
    #define CAT_COL_STUDIO "Color Studio"

    UI_FLOAT(CAT_COL_STUDIO, UI_CC_Exposure, "Exposure", "Changes the exposure of the scene", -5.0, 5.0, -0.25)
    UI_FLOAT2(CAT_COL_STUDIO, UI_CC_WBTempAndTint, "Temperature And Tint", "Changes the white balance of the image.", -0.5, 0.5, float2(-0.025, 0.025))
    UI_FLOAT(CAT_COL_STUDIO, UI_CC_Saturation, "Saturation", "Changes the saturation of the image", -1.0, 1.0, 0.05)
    UI_FLOAT(CAT_COL_STUDIO, UI_CC_Contrast, "Contrast", "Changes the contrast of the image", -1.0, 1.0, 0.05)
    UI_COLOR(CAT_COL_STUDIO, UI_CC_Lift, "Shadows", "Change the color of the shadows.", 0.0, 1.0, 0.5)
    UI_COLOR(CAT_COL_STUDIO, UI_CC_Gamma, "Midtones", "Change the color of the midtones.", 0.0, 1.0, 0.5)
    UI_COLOR(CAT_COL_STUDIO, UI_CC_Gain, "Highlights", "Changes the color of the highlights.", 0.0, 1.0, 0.5)
    UI_FLOAT(CAT_COL_STUDIO, UI_CC_LiftOffs, "Shadows Offset", "Change the luminance of the shadows.", -1.0, 1.0, 0.0)
    UI_FLOAT(CAT_COL_STUDIO, UI_CC_GammaOffs, "Midtones Offset", "Change the luminance of the midtones.", -1.0, 1.0, 0.0)
    UI_FLOAT(CAT_COL_STUDIO, UI_CC_GainOffs, "Highlights Offset", "Changes the luminance of the highlights.", -1.0, 1.0, 0.0)
#endif
